@@ -12,6 +12,9 @@ import json
 import functools
 from datetime import datetime
 from shutil import make_archive, unpack_archive
+from multiprocessing import Pool
+from functools import partial
+
 
 from cognite.client.exceptions import CogniteNotFoundError, CogniteAPIError
 
@@ -50,7 +53,7 @@ def set_up_cdf_client(context='run'):
         if k not in global_client:
             global_client[k] = get_client(api_keys[k], k)
         
-    logging.info("CDF client was set up correctly")
+    logging.debug("CDF client was set up correctly")
 
 
 def get_client(api_key, api_key_label=None):
@@ -276,13 +279,50 @@ def call_function(function_name, data):
     client = global_client["functions"]
     function = client.functions.retrieve(external_id=function_name)
     logging.debug(f"Retrieved function {function_name}")
-    call = function.call(data)
-    logging.debug(f"Called function (call_id={call.id})")
+    call_complete = False
+    t_start = time.process_time()
+    while not call_complete:
+        try:
+            call = function.call(data)
+            call_complete = True
+        except CogniteAPIError:
+            logging.warning(f"Cognite API error, try again ({call.id=})")
+    logging.debug(f"Called function ({call.id=})")
     response = call.get_response()
-    logging.debug(f"Called function: {response=}")
+    t_proc = time.process_time() - t_start
+    logging.debug(f"Call complete ({call.id=}, {t_proc=}): {response=}")
     return response
 
 
+def call_function_process_wrapper(function_name, data):
+    """
+    Set up the cdf client and call a function. This wrapper makes it possible to call a function from an independent processess invoked by the multiprocessing library
+    
+    Input:
+        - function_name: (string) function's external id in CDF
+        - data: (dictionary) data for the call
+    Output:
+        - (dictionary) function's response
+    """
+    set_up_cdf_client(context='deploy')
+    return call_function(function_name, data)
+
+
+def call_function_parallel(function_name, data, n_calls=None):
+    """
+    Make parallel calls to a function deployed in CDF
+
+    Input:
+        - function_name: (string) function's external id in CDF
+        - data: (list of dictionaries) list of data for the call
+        - n_calls: (optional int) number of parallel calls (set None to use all 
+            available cpu cores)
+    """
+    f = partial(call_function_process_wrapper, function_name)
+    with Pool(n_calls) as p:
+        return p.map(f, data)
+
+
 def test_function(function_name, data):
     """
     Call a function with data and verify that the response's 
